"""
Testing the evaluation functions.
"""

import os
import shutil
import glob

from sklearn import datasets

from smote_variants.base import equal_dicts

from smote_variants.evaluation import (evaluate_oversamplers, model_selection,
                                        execute_1_job, execute_1_os, execute_1_eval,
                                        do_clean_up, do_parse_results, Folding)

dataset_0 = datasets.load_breast_cancer()
dataset_0['name'] = 'breast_cancer'
dataset_1 = datasets.load_iris()
dataset_1['name'] = 'iris'
dataset_1['target'][dataset_1['target'] == 2] = 1

oversamplers = [('smote_variants', 'SMOTE', {'n_jobs': 1, 'random_state': 5}),
            ('smote_variants', 'SMOTE', {'n_jobs': 1, 'n_neighbors': 9, 'random_state': 5})]

classifiers = [('sklearn.tree', 'DecisionTreeClassifier', {'random_state': 5}),
                ('sklearn.neighbors', 'KNeighborsClassifier', {'n_jobs': 1})]

cache_path = os.path.join('.', 'test_path')

folding = Folding(dataset_0)

def test_execute_1_job():
    """
    Testing the execution of 1 evaluation job.
    """
    fold = next(folding.fold())
    result = execute_1_job(fold,
                            oversamplers,
                            ('sklearn.preprocessing', 'StandardScaler', {}),
                            classifiers,
                            serialization='json')
    assert len(result) > 0

def test_execute_1_subjob():
    """
    Testing the execution of 1 oversampling and 1 evaluation subjob.
    """
    fold = next(folding.fold())
    result = execute_1_os(fold,
                            oversamplers,
                            ('sklearn.preprocessing', 'StandardScaler', {}),
                            serialization='json',
                            cache_path=None)
    assert len(result) > 0

    result = execute_1_eval(result[0],
                            classifiers,
                            serialization='json',
                            cache_path=None)
    assert len(result) > 0

def test_do_clean_up():
    """
    Testing the clean-up functionalities.
    """
    os.makedirs(cache_path, exist_ok=True)

    with open(os.path.join(cache_path, 'fold_dummy.txt'),
                'wt', encoding='utf-8') as file:
        file.write('dummy')

    with open(os.path.join(cache_path, 'oversampling_dummy.txt'),
                'wt', encoding='utf-8') as file:
        file.write('dummy')

    with open(os.path.join(cache_path, 'evaluation_dummy.txt'),
                'wt', encoding='utf-8') as file:
        file.write('dummy')

    do_clean_up(cache_path, 'oversamplings')

    assert len(glob.glob(os.path.join(cache_path, 'oversampling*'))) == 0
    assert len(glob.glob(os.path.join(cache_path, 'fold*'))) > 0
    assert len(glob.glob(os.path.join(cache_path, 'evaluation*'))) > 0

    do_clean_up(cache_path, 'all')

    assert len(glob.glob(os.path.join(cache_path, 'oversampling*'))) == 0
    assert len(glob.glob(os.path.join(cache_path, 'fold*'))) == 0
    assert len(glob.glob(os.path.join(cache_path, 'evaluation*'))) > 0

    shutil.rmtree(cache_path)

def test_evalute_oversamplers():
    """
    Testing the evaluation of oversamplers.
    """
    results_0 = evaluate_oversamplers(datasets=[dataset_0, dataset_1],
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=None,
            validator_params={'n_repeats': 2, 'n_splits': 5, 'random_state': 5},
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            parse_results=True)

    shutil.rmtree(cache_path, ignore_errors=True)

    results_1 = evaluate_oversamplers(datasets=[dataset_0, dataset_1],
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=cache_path,
            validator_params={'n_repeats': 2, 'n_splits': 5, 'random_state': 5},
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            parse_results=False)

    results_1 = do_parse_results(results_1, False, serialization='json')
    results_1 = do_parse_results(results_1, True, serialization='json')

    shutil.rmtree(cache_path, ignore_errors=True)

    assert results_0['acc_mean'].max() == results_1['acc_mean'].max()

def test_model_selection():
    """
    Testing the model selection.
    """
    samp_0, clas_0 = model_selection(dataset=dataset_0,
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=None,
            validator_params={'n_repeats': 2, 'n_splits': 5, 'random_state': 5},
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            score='auc')

    shutil.rmtree(cache_path, ignore_errors=True)

    samp_1, clas_1 = model_selection(dataset=dataset_0,
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=cache_path,
            validator_params=None,
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            score='auc')

    shutil.rmtree(cache_path, ignore_errors=True)

    assert equal_dicts(samp_0.get_params(), samp_1.get_params())
    assert equal_dicts(clas_0.get_params(), clas_1.get_params())



"""
Testing the CURE_SMOTE.
"""

import numpy as np

from smote_variants import CURE_SMOTE

from smote_variants.datasets import (load_all_min_noise,
                                     load_3_min_some_maj)

def test_specific():
    """
    Oversampler specific testing
    """

    obj = CURE_SMOTE(noise_th=10)

    dataset = load_all_min_noise()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0

    obj = CURE_SMOTE(n_clusters=3)

    dataset = load_3_min_some_maj()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0

    X_min = np.array([[1.0, 2.0], [2.0, 3.0], [3.0, 4.0]])
    clusters = [np.array([0]), np.array([1]), np.array([2])]
    samples = obj.generate_samples_in_clusters(X_min, clusters, 10)

    assert len(samples) > 0



"""
This module implements the Safe_Level_SMOTE method.
"""

import numpy as np

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['Safe_Level_SMOTE']

class Safe_Level_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @inproceedings{safe_level_smote,
                        author = {
                            Bunkhumpornpat, Chumphol and Sinapiromsaran,
                        Krung and Lursinsap, Chidchanok},
                        title = {Safe-Level-SMOTE: Safe-Level-Synthetic
                                Minority Over-Sampling TEchnique for
                                Handling the Class Imbalanced Problem},
                        booktitle = {Proceedings of the 13th Pacific-Asia
                                    Conference on Advances in Knowledge
                                    Discovery and Data Mining},
                        series = {PAKDD '09},
                        year = {2009},
                        isbn = {978-3-642-01306-5},
                        location = {Bangkok, Thailand},
                        pages = {475--482},
                        numpages = {8},
                        url = {http://dx.doi.org/10.1007/978-3-642-01307-2_43},
                        doi = {10.1007/978-3-642-01307-2_43},
                        acmid = {1533904},
                        publisher = {Springer-Verlag},
                        address = {Berlin, Heidelberg},
                        keywords = {Class Imbalanced Problem, Over-sampling,
                                    SMOTE, Safe Level},
                    }

    Notes:
        * The original method was not prepared for the case when no minority
            sample has minority neighbors.
    """

    categories = [OverSampling.cat_borderline,
                  OverSampling.cat_extensive,
                  OverSampling.cat_sample_componentwise,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params= {},
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): control parameter of the nearest neighbor
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1.0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def slp_sln(self, y, indices):
        """
        Creating an array of allowed positive negative pairs
        and the corresponding safe level scores.

        Args:
            y (np.array): the target labels
            indices (np.array): the neighborhood structure

        Returns:
            np.array: an array with the first column containing the positive
                        index, the second the neighbor index, the third and
                        fourth the corresponding safe levels
        """
        minority_indices = np.where(y == self.min_label)[0]

        safe_levels = np.sum(y[indices[:, 1:]] == self.min_label, axis=1)

        pos_slp = minority_indices[safe_levels[minority_indices] > 0]
        null_slp = minority_indices[safe_levels[minority_indices] == 0]

        part0 = np.vstack([np.repeat(pos_slp, indices.shape[1] - 1),
                            indices[pos_slp, 1:].flatten()]).T

        pos_sln = np.where(safe_levels[indices[null_slp, 1:]] > 0)
        part1 = np.vstack([null_slp[pos_sln[0]],
                            indices[null_slp, 1:][pos_sln]]).T

        pairs = np.vstack([part0, part1])
        return np.vstack([pairs.T, safe_levels[pairs[:, 0]], safe_levels[pairs[:, 1]]]).T

    def generate_samples(self, X, y, slp_sln, n_to_sample):
        """
        Generate samples.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            slp_sln (np.array): the slp-sln matrix
            n_to_smaple (int): the number of samples to generate

        Returns:
            np.array, np.array: the extended training set
        """
        if len(slp_sln) == 0:
            return self.return_copies(X, y, "No neighbors passing the "\
                                                            "safe level test")

        base_indices = \
            self.random_state.choice(np.arange(len(slp_sln)), n_to_sample)

        gap = self.random_state.random_sample(size=(n_to_sample, X.shape[1]))

        slp = slp_sln[base_indices][:,2]
        sln = slp_sln[base_indices][:,3]

        gap = (gap.T * np.where(np.logical_and(sln == 0, slp > 0), 0.0, 1.0)).T

        mask = slp > sln
        gap[mask] = (gap[mask].T * sln[mask]/slp[mask]).T

        mask = slp < sln
        gap[mask] = (gap[mask].T * slp[mask]/sln[mask]).T
        gap[mask] = (gap[mask].T + (1.0 - slp[mask]/sln[mask])).T

        samples = X[slp_sln[base_indices][:,0]] + \
                    ((X[slp_sln[base_indices][:,1]] - X[slp_sln[base_indices][:,0]]) * gap)

        return (np.vstack([X, samples]),
               np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # fitting nearest neighbors model
        n_neighbors = min([self.n_neighbors+1, len(X)])

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        indices = nnmt.kneighbors(X, return_distance=False)

        slp_sln = self.slp_sln(y, indices)

        return self.generate_samples(X, y, slp_sln, n_to_sample)


    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the E_SMOTE method.
"""

import numpy as np

from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['E_SMOTE']

class E_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{e_smote,
                            author={Deepa, T. and Punithavalli, M.},
                            booktitle={2011 3rd International Conference on
                                        Electronics Computer Technology},
                            title={An E-SMOTE technique for feature selection
                                    in High-Dimensional Imbalanced Dataset},
                            year={2011},
                            volume={2},
                            number={},
                            pages={322-324},
                            keywords={bioinformatics;data mining;pattern
                                        classification;support vector machines;
                                        E-SMOTE technique;feature selection;
                                        high-dimensional imbalanced dataset;
                                        data mining;bio-informatics;dataset
                                        balancing;SVM classification;micro
                                        array dataset;Feature extraction;
                                        Genetic algorithms;Support vector
                                        machines;Data mining;Machine learning;
                                        Bioinformatics;Cancer;Imbalanced
                                        dataset;Featue Selection;E-SMOTE;
                                        Support Vector Machine[SVM]},
                            doi={10.1109/ICECTECH.2011.5941710},
                            ISSN={},
                            month={April}}

    Notes:
        * This technique is basically unreproducible. I try to implement
            something following the idea of applying some simple genetic
            algorithm for optimization.
        * In my best understanding, the technique uses evolutionary algorithms
            for feature selection and then applies vanilla SMOTE on the
            selected features only.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_dim_reduction,
                  OverSampling.cat_memetic,
                  OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 min_features=2,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in the nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            min_features (int): minimum number of features
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(min_features, "min_features", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.min_features = min_features
        self.n_jobs = n_jobs

        self.mask = None

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'min_features': [1, 2, 3]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def crossover(self, mask_a, mask_b, min_features):
        """
        Crossover operation for two masks

        Args:
            mask_a (np.array): binary mask 1
            mask_b (np.array): binary mask 2

        Returns:
            np.array: the result of crossover
        """
        mask = mask_a.copy()
        for idx, flag in enumerate(mask_b):
            if self.random_state.randint(0, 2) == 0:
                mask[idx] = flag

        while np.sum(mask) < min_features:
            mask[self.random_state.randint(len(mask))] = True

        return mask

    def mutate(self, mask_old, min_features):
        """
        Mutation operation for a mask

        Args:
            mask_old (np.array): binary mask

        Returns:
            np.array: the result of mutation
        """
        mask = mask_old.copy()
        for idx, flag in enumerate(mask):
            if self.random_state.randint(0, 2) == 0:
                mask[idx] = not flag

        while np.sum(mask) < min_features:
            mask[self.random_state.randint(len(mask))] = True

        return mask

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        min_features = min(self.min_features, len(X[0]))

        if len(X) < 800:
            classifier = SVC(gamma='auto', random_state=self._random_state_init)
        else:
            classifier = DecisionTreeClassifier(
                max_depth=4, random_state=self._random_state_init)

        # parameters of the evolutionary algorithm
        n_generations = 50
        n_population = 5

        # creating initial mask
        mask = np.repeat([False], X.shape[1])
        mask[np.random.choice(np.arange(mask.shape[0]), 1)] = True

        # generating initial population
        population = [[0, mask.copy()] for _ in range(n_population)]
        for _ in range(n_generations):
            # in each generation
            for _ in range(n_population):
                # for each element of a population
                if self.random_state.randint(0, 2) == 0:
                    # crossover
                    i_0 = self.random_state.randint(n_population)
                    i_1 = self.random_state.randint(n_population)
                    mask = self.crossover(population[i_0][1],
                                            population[i_1][1],
                                            min_features)
                else:
                    # mutation
                    idx = self.random_state.randint(n_population)
                    mask = self.mutate(population[idx][1],
                                        min_features)
                # evaluation
                _logger.info("%s evaluting mask selection with features %d/%d",
                            self.__class__.__name__, np.sum(mask), len(mask))
                classifier.fit(X[:, mask], y)
                score = np.sum(y == classifier.predict(X[:, mask]))/len(y)
                # appending the result to the population
                population.append([score, mask])
            # sorting the population in a reversed order and keeping the
            # elements with the highest scores
            population = sorted(population, key=lambda x: -x[0])[:n_population]

        self.mask = population[0][1]
        # resampling the population in the given dimensions

        smote = SMOTE(proportion=self.proportion,
                      n_neighbors=self.n_neighbors,
                      nn_params=self.nn_params,
                      ss_params=self.ss_params,
                      n_jobs=self.n_jobs,
                      random_state=self._random_state_init)

        return smote.sample(X[:, self.mask], y)

    def preprocessing_transform(self, X):
        """
        Transform new data by the learnt transformation

        Args:
            X (np.array): new data

        Returns:
            np.array: transformed data
        """
        return X[:, self.mask]

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'min_features': self.min_features,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the SOI_CJ.
"""

from smote_variants import SOI_CJ

from smote_variants.datasets import load_illustration_2_class

def test_specific():
    """
    Oversampler specific testing
    """
    dataset = load_illustration_2_class()

    obj = SOI_CJ(method='interpolation')

    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    obj = SOI_CJ(method='jittering')

    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    cluster_idx = set([1, 2, 3])
    cluster_jdx = set([3, 4, 5, 6, 7])
    intersection = set([3])

    cluster_idx, cluster_jdx = obj.rearrange_clusters(cluster_idx,
                                                      cluster_jdx,
                                                      intersection)

    assert len(cluster_jdx) > 0



"""
Tests oversamplers with equalized data.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_same_num

dataset = load_same_num()

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_same_num(smote_class):
    """
    Tests oversamplers with equalized data.

    Args:
        smote_class (class): an oversampling class
    """
    X, y = smote_class().sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
This module implements the SMOBD method.
"""
import warnings

import numpy as np

from sklearn.cluster import OPTICS

from ..base import fix_density
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['SMOBD']

class SMOBD(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{smobd,
                            author={Cao, Q. and Wang, S.},
                            booktitle={2011 International Conference on
                                        Information Management, Innovation
                                        Management and Industrial
                                        Engineering},
                            title={Applying Over-sampling Technique Based
                                     on Data Density and Cost-sensitive
                                     SVM to Imbalanced Learning},
                            year={2011},
                            volume={2},
                            number={},
                            pages={543-548},
                            keywords={data handling;learning (artificial
                                        intelligence);support vector machines;
                                        oversampling technique application;
                                        data density;cost sensitive SVM;
                                        imbalanced learning;SMOTE algorithm;
                                        data distribution;density information;
                                        Support vector machines;Classification
                                        algorithms;Noise measurement;Arrays;
                                        Noise;Algorithm design and analysis;
                                        Training;imbalanced learning;
                                        cost-sensitive SVM;SMOTE;data density;
                                        SMOBD},
                            doi={10.1109/ICIII.2011.276},
                            ISSN={2155-1456},
                            month={Nov},}
    """

    categories = [OverSampling.cat_uses_clustering,
                  OverSampling.cat_density_based,
                  OverSampling.cat_extensive,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 eta1=0.5,
                 t=1.8,
                 min_samples=5,
                 nn_params={},
                 max_eps=1.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            eta1 (float): control parameter of density estimation
            t (float): control parameter of noise filtering
            min_samples (int): minimum samples parameter for OPTICS
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            max_eps (float): maximum environment radius parameter for OPTICS
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_in_range(eta1, 'eta1', [0.0, 1.0])
        self.check_greater_or_equal(t, 't', 0)
        self.check_greater_or_equal(min_samples, 'min_samples', 1)
        self.check_greater_or_equal(max_eps, 'max_eps', 0.0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.eta1 = eta1
        self.t = t # pylint: disable=invalid-name
        self.min_samples = min_samples
        self.nn_params = nn_params
        self.max_eps = max_eps
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'eta1': [0.1, 0.5, 0.9],
                                  't': [1.5, 2.5],
                                  'min_samples': [5],
                                  'max_eps': [0.1, 0.5, 1.0, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_indices_and_density(self, X_min, nn_params):
        """
        Determine the neighborhood structure and the density

        Args:
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array, np.array: the neighborhood structure and the density
        """
        # running the OPTICS technique based on the sklearn implementation
        min_samples = min([len(X_min)-1, self.min_samples])
        optics = OPTICS(min_samples=min_samples,
                   max_eps=self.max_eps,
                   n_jobs=self.n_jobs)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            optics.fit(X_min)

        # noise filtering
        cd_average = np.mean(optics.core_distances_)
        rd_average = np.mean(optics.reachability_)
        noise = np.logical_and(optics.core_distances_ > cd_average*self.t,
                                optics.reachability_ > rd_average*self.t)

        # fitting a nearest neighbor model to be able to find
        # neighbors in radius
        n_neighbors = min([len(X_min), self.min_samples+1])

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        indices = nnmt.kneighbors(X_min, return_distance=False)

        # determining the density
        factor_1 = optics.core_distances_
        factor_2 = np.array([len(x)
                        for x in nnmt.radius_neighbors(X_min,
                                                       radius=self.max_eps,
                                                       return_distance=False)])

        if np.all(factor_1 == np.inf) or np.all(factor_2 == 0):
            raise ValueError("factor_1 or factor_2 is not suitable")

        factor_1[factor_1 == np.inf] = max(factor_1[factor_1 != np.inf])*1.1

        factor_1 = factor_1 / max([max(factor_1), 1e-8])
        factor_2 = factor_2 / max([max(factor_2), 1e-8])

        density = factor_1*self.eta1 + factor_2*(1 - self.eta1)

        # setting the density at noisy samples to zero
        density[noise] = 0.0

        density = fix_density(density)

        return indices, density

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        try:
            indices, density = self.determine_indices_and_density(X_min,
                                                                  nn_params)
        except ValueError as valueerror:
            return self.return_copies(X, y, valueerror.args[0])

        base_ind = self.random_state.choice(np.arange(X_min.shape[0]),
                                            n_to_sample,
                                            p=density)
        neigh_ind = self.random_state.choice(np.arange(1, indices.shape[1]),
                                             n_to_sample)
        base_vectors = X_min[base_ind]
        neigh_vectors = X_min[indices[base_ind, neigh_ind]]
        random = self.random_state.random_sample(size=base_vectors.shape)

        samples = base_vectors + (neigh_vectors - base_vectors) *  random

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'eta1': self.eta1,
                't': self.t,
                'min_samples': self.min_samples,
                'nn_params': self.nn_params,
                'max_eps': self.max_eps,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the kmeans_SMOTE method.
"""
import warnings

import numpy as np

from sklearn.cluster import KMeans

from ..base import coalesce, coalesce_dict, fix_density
from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['kmeans_SMOTE']

class kmeans_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{kmeans_smote,
                        title = "Improving imbalanced learning through a
                                    heuristic oversampling method based
                                    on k-means and SMOTE",
                        journal = "Information Sciences",
                        volume = "465",
                        pages = "1 - 20",
                        year = "2018",
                        issn = "0020-0255",
                        doi = "https://doi.org/10.1016/j.ins.2018.06.056",
                        author = "Georgios Douzas and Fernando Bacao and
                                    Felix Last",
                        keywords = "Class-imbalanced learning, Oversampling,
                                    Classification, Clustering, Supervised
                                    learning, Within-class imbalance"
                        }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_clusters=10,
                 irt=2.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_clusters (int): number of clusters
            irt (float): imbalanced ratio threshold
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(n_clusters, "n_clusters", 1)
        self.check_greater_or_equal(irt, "irt", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.n_clusters = n_clusters
        self.irt = irt
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_clusters': [2, 5, 10, 20, 50],
                                  'irt': [0.5, 0.8, 1.0, 1.5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def do_the_clustering(self, X, y):
        """
        Do the clustering.

        Args:
            X (np.array): all feature vectors
            y (np.array): all target labels

        Returns:
            list: the clusters
        """
        # applying kmeans clustering to all data
        n_clusters = min([self.n_clusters, len(X)])
        kmeans = KMeans(n_clusters=n_clusters,
                        random_state=self._random_state_init)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            kmeans.fit(X)

        # extracting clusters
        labels = kmeans.labels_
        clusters = [np.where(labels == li)[0] for li in range(n_clusters)]

        # cluster filtering
        def cluster_filter(cluster):
            numerator = np.sum(y[cluster] == self.maj_label) + 1
            denominator = np.sum(y[cluster] == self.min_label) + 1
            n_minority = np.sum(y[cluster] == self.min_label)
            return numerator/denominator < self.irt and n_minority > 1

        filt_clusters = [cluster for cluster in clusters \
                                        if cluster_filter(cluster)]

        return filt_clusters

    def calculate_weights(self, filt_clusters, X, y):
        """
        Calculate the weights.

        Args:
            filt_clusters (list): the filtered clusters
            X (np.array): all feature vectors
            y (np.array): all target labels

        Returns:
            np.array, list, list: the weights, the cluster minority indices,
                                    the nearest neighbors
        """
        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        # Step 2 in the paper
        sparsity = []
        nearest_neighbors = []
        cluster_minority_ind = []
        for cluster in filt_clusters:
            # extract minority indices in the cluster
            minority_ind = cluster[y[cluster] == self.min_label]
            cluster_minority_ind.append(minority_ind)
            # compute distance matrix of minority samples in the cluster
            distm = pairwise_distances_mahalanobis(X[minority_ind],
                                    tensor=nn_params['metric_tensor'])
            min_count = len(minority_ind)
            # compute the average of distances
            avg_min_dist = (np.sum(distm) - distm.trace()) / \
                (len(minority_ind)**2 - len(minority_ind))
            # compute sparsity (Step 4)
            sparsity.append(avg_min_dist**len(X[0])/min_count)
            # extract the nearest neighbors graph
            n_neighbors = min([len(minority_ind), self.n_neighbors + 1])
            nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
            nnmt.fit(X[minority_ind])
            nearest_neighbors.append(nnmt.kneighbors(X[minority_ind]))

        # Step 5 - compute density of sampling
        sparsity = fix_density(sparsity)

        return sparsity, cluster_minority_ind, nearest_neighbors

    def generate_samples_in_clusters(self,
                                    *,
                                    X,
                                    weights,
                                    cluster_minority_ind,
                                    nearest_neighbors,
                                    n_to_sample):
        """
        Generates samples within the clusters.

        Args:
            X (np.array): all training vectors
            weights (np.array): the cluster weights
            cluster_minority_ind (list): minority indices corresponding
                                            to the clusters
            nearest neighbors (list): the nearest neighbor indices
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """

        clusters_selected = self.random_state.choice(len(weights),
                                                    n_to_sample,
                                                    p=weights)
        cluster_unique, cluster_count = np.unique(clusters_selected,
                                                    return_counts=True)

        #n_dim_original = self.n_dim
        samples = []
        for idx, cluster in enumerate(cluster_unique):
            cluster_vectors = X[cluster_minority_ind[cluster]]
            #self.n_dim = np.min([self.n_dim, cluster_vectors.shape[0]])
            samples.append(self.sample_simplex(X=cluster_vectors,
                                        indices=nearest_neighbors[cluster][1],
                                        n_to_sample=cluster_count[idx]))
            #self.n_dim = n_dim_original

        return np.vstack(samples)

        # do the sampling
        #samples = []
        #while len(samples) < n_to_sample:
        #    # choose random cluster index and random minority element
        #    clust_ind = self.random_state.choice(
        #        np.arange(len(weights)), p=weights)
        #    idx = self.random_state.randint(
        #        len(cluster_minority_ind[clust_ind]))
        #    base_idx = cluster_minority_ind[clust_ind][idx]
        #    # choose random neighbor
        #    neighbor_cluster_indices = nearest_neighbors[clust_ind][1][idx][1:]
        #    domain = cluster_minority_ind[clust_ind][neighbor_cluster_indices]
        #    neighbor_idx = self.random_state.choice(domain)
        #    # sample
        #    X_a = X[base_idx]
        #    X_b = X[neighbor_idx]
        #    samples.append(self.sample_between_points(X_a, X_b))

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        filt_clusters = self.do_the_clustering(X, y)

        if len(filt_clusters) == 0:
            return self.return_copies(X, y, "number of clusters after "\
                                                            "filtering is 0")

        weights, cluster_minority_ind, nearest_neighbors = \
                    self.calculate_weights(filt_clusters, X, y)

        samples = self.generate_samples_in_clusters(X=X,
                                    weights=weights,
                                    cluster_minority_ind=cluster_minority_ind,
                                    nearest_neighbors=nearest_neighbors,
                                    n_to_sample=n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_clusters': self.n_clusters,
                'irt': self.irt,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing oversampling with simplex sampling dim 1.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_illustration_2_class

dataset = load_illustration_2_class()

@pytest.mark.parametrize("smote_class",
    sv.get_simplex_sampling_oversamplers())
def test_simplex_dim_1(smote_class):
    """
    Testing oversamplers with simplex sampling dim 1.

    Args:
        smote_class (class): an oversampler class.
    """
    ss_params = {'n_dim': 1}
    X, y = smote_class(ss_params=ss_params).sample(dataset['data'],
                                                   dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
A dummy classifier raising errors and warnings.
"""

import warnings

from sklearn.neighbors import KNeighborsClassifier

__all__=['ErrorWarningClassifier']

class ErrorWarningClassifier:
    """
    Mock classifier throwing an error or warning.
    """

    def __init__(self, raise_value_error=False,
                       raise_warning=False,
                       raise_runtime_error=False):
        """
        Constructor of the classifier.

        Args:
            raise_value_error (bool): whether to raise a ValueError
            raise_warning (bool): whether to raise a warning
            raise_runtime_error (bool): whether to raise a RuntimeError
        """
        self.raise_value_error = raise_value_error
        self.raise_warning = raise_warning
        self.raise_runtime_error = raise_runtime_error
        self.classifier = KNeighborsClassifier()

    def fit(self, X, y):
        """
        Fitting the classifier

        Args:
            X (np.array): the training vectors
            y (np.array): the target labels

        Returns:
            obj: the fitted object
        """

        if self.raise_value_error:
            raise ValueError("Dummy value error")

        if self.raise_runtime_error:
            raise RuntimeError("Dummy runtime error")

        if self.raise_warning:
            warnings.warn("Dummy warning")

        self.classifier.fit(X, y)
        return self

    def predict(self, X):
        """
        Predict the new vectors.

        Args:
            X (np.array): new vectors to predict

        Returns:
            np.array: the predicted labels
        """

    def predict_proba(self, X):
        """
        Predict probabilities for the new vectors.

        Args:
            X (np.array): new vectors to predict

        Returns:
            np.array: the predicted probabilities
        """
        return self.classifier.predict_proba(X)

    def get_params(self):
        """
        Return the parameters

        Returns:
            dict: the parameters
        """
        return {'raise_value_error': self.raise_value_error,
                'raise_warning': self.raise_warning,
                'raise_runtime_error': self.raise_runtime_error}



"""
This module implements the MDO method.
"""

import numpy as np

from sklearn.decomposition import PCA

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['MDO']

class MDO(OverSampling):
    """
    References:
        * BibTex::

            @ARTICLE{mdo,
                        author={Abdi, L. and Hashemi, S.},
                        journal={IEEE Transactions on Knowledge and Data
                                    Engineering},
                        title={To Combat Multi-Class Imbalanced Problems
                                by Means of Over-Sampling Techniques},
                        year={2016},
                        volume={28},
                        number={1},
                        pages={238-251},
                        keywords={covariance analysis;learning (artificial
                                    intelligence);modelling;pattern
                                    classification;sampling methods;
                                    statistical distributions;minority
                                    class instance modelling;probability
                                    contour;covariance structure;MDO;
                                    Mahalanobis distance-based oversampling
                                    technique;data-oriented technique;
                                    model-oriented solution;machine learning
                                    algorithm;data skewness;multiclass
                                    imbalanced problem;Mathematical model;
                                    Training;Accuracy;Eigenvalues and
                                    eigenfunctions;Machine learning
                                    algorithms;Algorithm design and analysis;
                                    Benchmark testing;Multi-class imbalance
                                    problems;over-sampling techniques;
                                    Mahalanobis distance;Multi-class imbalance
                                    problems;over-sampling techniques;
                                    Mahalanobis distance},
                        doi={10.1109/TKDE.2015.2458858},
                        ISSN={1041-4347},
                        month={Jan}}
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_dim_reduction,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 K2=5,
                 K1_frac=0.5,
                 nn_params={},
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            K2 (int): number of neighbors
            K1_frac (float): the fraction of K2 to set K1
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(K2, "K2", 1)
        self.check_greater_or_equal(K1_frac, "K1_frac", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.K2 = K2 # pylint: disable=invalid-name
        self.K1_frac = K1_frac # pylint: disable=invalid-name
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'K2': [3, 5, 7],
                                  'K1_frac': [0.3, 0.5, 0.7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_1_sample(self,
                            alpha,
                            V # pylint: disable=invalid-name
                            ):
        """
        Generate 1 sample.

        Args:
            alpha (float): the alpha value
            V (np.array): the variances

        Returns:
            np.array: the generated sample
        """
        alpha_V = alpha * V # pylint: disable=invalid-name
        alpha_V[alpha_V < 0.001] = 0.001

        # initializing a new vector
        X_new = np.zeros(V.shape[0])

        # sampling components of the new vector
        s = 0# pylint: disable=invalid-name
        for jdx in range(V.shape[0] - 1):
            r = (2 * self.random_state.random_sample() - 1) # pylint: disable=invalid-name
            r = r * np.sqrt(alpha_V[jdx]) # pylint: disable=invalid-name
            X_new[jdx] = r
            s = s + (r**2 / alpha_V[jdx]) # pylint: disable=invalid-name

        last_fea_val = np.sqrt(max((1 - s) * alpha * V[-1], 0)) * (s <= 1)

        # determine last component to fulfill the ellipse equation
        X_new[-1] = (2 * self.random_state.random_sample() - 1) \
                                                        * last_fea_val

        return X_new

    def generate_samples(self, *, X_sel, weights, n_to_sample):
        """
        Generate samples.

        Args:
            X_sel (np.array): the selected samples
            weights (np.array): the weights
            n_to_sample (int): the number of samples

        Returns:
            np.array: the generated samples
        """
        # Algorithm 1 - MDO over-sampling
        mu = np.mean(X_sel, axis=0) # pylint: disable=invalid-name
        Z = X_sel - mu # pylint: disable=invalid-name
        # executing PCA
        pca = PCA(n_components=min([Z.shape[1], Z.shape[0]])).fit(Z)
        T = pca.transform(Z) # pylint: disable=invalid-name

        # computing variances (step 13)
        V = np.var(T, axis=0) # pylint: disable=invalid-name

        V[V < 0.001] = 0.001

        # generating samples
        samples = []
        while len(samples) < n_to_sample:
            # selecting a sample randomly according to the distribution
            idx = self.random_state.choice(np.arange(X_sel.shape[0]),
                                            p=weights)

            # finding vector in PCA space

            # computing alpha
            alpha = np.sum(T[idx]**2 / V)

            X_new = self.generate_1_sample(alpha, V)

            # append to new samples
            samples.append(X_new)

        return pca.inverse_transform(np.vstack(samples)) + mu

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # determining K1
        K1 = int(self.K2*self.K1_frac) # pylint: disable=invalid-name
        K1 = min([K1, len(X)]) # pylint: disable=invalid-name
        K2 = min([self.K2 + 1, len(X)]) # pylint: disable=invalid-name

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        # Algorithm 2 - chooseSamples
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=K2,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        # extracting the number of minority samples in local neighborhoods
        n_min = np.sum(y[ind[:, 1:]] == self.min_label, axis=1)

        # extracting selected samples from minority ones
        X_sel = X_min[n_min >= K1] # pylint: disable=invalid-name

        # falling back to returning input data if all the input is considered
        # noise
        if X_sel.shape[0] == 0:
            return self.return_copies(X, y, "No samples selected")

        # computing distribution
        weights = n_min[n_min >= K1] / K2
        weights = weights/np.sum(weights)

        samples = self.generate_samples(X_sel=X_sel,
                                        weights=weights,
                                        n_to_sample=n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'K2': self.K2,
                'K1_frac': self.K1_frac,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the VIS_RST method.
"""

import numpy as np

from sklearn.preprocessing import StandardScaler

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['VIS_RST']

class VIS_RST(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{vis_rst,
                            author="Borowska, Katarzyna
                            and Stepaniuk, Jaroslaw",
                            editor="Saeed, Khalid
                            and Homenda, Wladyslaw",
                            title="Imbalanced Data Classification: A Novel
                                    Re-sampling Approach Combining Versatile
                                    Improved SMOTE and Rough Sets",
                            booktitle="Computer Information Systems and
                                        Industrial Management",
                            year="2016",
                            publisher="Springer International Publishing",
                            address="Cham",
                            pages="31--42",
                            isbn="978-3-319-45378-1"
                            }

    Notes:
        * Replication of DANGER samples will be removed by the last step of
            noise filtering.
        * The rules in the paper do not cover all cases
    """

    categories = [OverSamplingSimplex.cat_changes_majority,
                  OverSamplingSimplex.cat_noise_removal,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0.0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def relabel_boundary(self, X, y, nn_params):
        """
        Relabel the boundary vectors

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array: the new minority samples and the updated
                                target labels
        """
        # fitting nearest neighbors model to determine boundary region
        n_neighbors = min([len(X), self.n_neighbors + 1])

        X_maj = X[y == self.maj_label]

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        ind = nnmt.kneighbors(X_maj, return_distance=False)

        # determining boundary region of majority samples
        boundary = np.sum(y[ind] == self.maj_label, axis=1) != n_neighbors

        y_maj = y[y == self.maj_label]
        y_maj[boundary] = self.min_label
        y[y == self.maj_label] = y_maj

        # extracting new minority and majority set
        X_min = X[y == self.min_label]

        return X_min, y

    def label_minority_samples(self, X, y, X_min, nn_params):
        """
        Label the minority samples.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            X_min (np.array): the minority vectors
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the labels
        """
        # fitting nearest neighbors model to determine boundary region
        n_neighbors = min([len(X), self.n_neighbors + 1])

        # labeling minority samples
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        indices = nnmt.kneighbors(X_min, return_distance=False)

        # extracting labels
        labels = np.array(['DAN'] * X_min.shape[0])

        min_class_neighbors = \
                    np.sum(y[indices[:, 1:]] == self.maj_label, axis=1)

        labels[min_class_neighbors == (n_neighbors - 1)] = 'NOI'
        labels[min_class_neighbors < (n_neighbors - 1) / 2] = 'SAF'

        return labels

    def set_mode(self, X_min, labels):
        """
        Determine the mode of oversampling.

        Args:
            X_min (np.array): the minority samples
            labels (np.array): the labels

        Returns:
            str: the mode
        """
        # extracting the number of different labels (noise is not used)
        safe = np.sum(labels == 'SAF')
        danger = np.sum(labels == 'DAN')

        if safe == 0:
            mode = 'no_safe'
        elif danger > 0.3*len(X_min):
            mode = 'high_complexity'
        else:
            mode = 'low_complexity'

        _logger.info("%s: selected mode: %s", self.__class__.__name__, mode)

        return mode

    def sampling_high_complexity(self, X_min, labels, ind_min, n_to_sample):
        """
        Implementation of the sampling rules for the high complexity case.

        Args:
            X_min (np.array): the minority samples
            labels (np.array): the minority labels
            ind_min (np.array): the neighborhood structure
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        # do the sampling
        not_noise_mask = labels != 'NOI'
        X_min_not_noise = X_min[not_noise_mask] # pylint: disable=invalid-name
        danger_mask = labels[not_noise_mask] == 'DAN'
        ind_not_noise = ind_min[not_noise_mask]

        samples = np.zeros((0, X_min.shape[1]))

        while samples.shape[0] < n_to_sample:
            n_missing = np.max([n_to_sample - samples.shape[0], 10])
            base_indices = self.random_state.choice(X_min_not_noise.shape[0],
                                                    n_missing)
            base_danger = base_indices[danger_mask[base_indices]]
            samples = np.vstack([samples, X_min_not_noise[base_danger]])

            base_not_danger = base_indices[~danger_mask[base_indices]]

            n_missing = base_not_danger.shape[0]

            base_not_danger = ~danger_mask

            samples_tmp = self.sample_simplex(X=X_min_not_noise[base_not_danger],
                                          indices=ind_not_noise[base_not_danger, 1:],
                                          n_to_sample=n_missing,
                                          X_vertices=X_min)

            samples = np.vstack([samples, samples_tmp])

        samples = samples[:n_to_sample]

        return samples

    def sampling_low_complexity(self, X_min, labels, ind_min, n_to_sample):
        """
        Implementation of the sampling rules for the high complexity case.

        Args:
            X_min (np.array): the minority samples
            labels (np.array): the minority labels
            ind_min (np.array): the neighborhood structure
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """

        # do the sampling
        not_noise_mask = labels != 'NOI'
        X_min_not_noise = X_min[not_noise_mask] # pylint: disable=invalid-name
        danger_mask = labels[not_noise_mask] == 'DAN'
        ind_not_noise = ind_min[not_noise_mask]

        samples = np.zeros((0, X_min.shape[1]))

        while samples.shape[0] < n_to_sample:
            n_missing = np.max([n_to_sample - samples.shape[0], 10])
            base_indices = self.random_state.choice(X_min_not_noise.shape[0],
                                                    n_missing)
            base_not_danger = base_indices[~danger_mask[base_indices]]
            samples = np.vstack([samples, X_min_not_noise[base_not_danger]])

            base_danger = base_indices[danger_mask[base_indices]]

            n_missing = base_danger.shape[0]

            base_danger = danger_mask

            samples_tmp = self.sample_simplex(X=X_min_not_noise[base_danger],
                                          indices=ind_not_noise[base_danger, 1:],
                                          n_to_sample=n_missing,
                                          X_vertices=X_min)

            samples = np.vstack([samples, samples_tmp])

        samples = samples[:n_to_sample]

        return samples

    def sampling_otherwise(self, X_min, ind_min, n_to_sample):
        """
        Implementation of the sampling rules for the high complexity case.

        Args:
            X_min (np.array): the minority samples
            ind_min (np.array): the neighborhood structure
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """

        samples = self.sample_simplex(X=X_min,
                                        indices=ind_min,
                                        n_to_sample=n_to_sample)

        return samples

    def generate_samples(self, *, X_min, nn_params, n_to_sample, labels, mode):
        """
        Generate samples.

        Args:
            X_min (np.array): all minority vectors
            nn_params (dict): the nearest neighbor parameters
            n_to_sample (int): the number of samples to generate
            labels (np.array): the labels of the minority samples
            mode (str): the sampling mode

        Returns:
            np.array: the generated samples
        """
        # fitting nearest neighbors to find the neighbors of minority elements
        # among minority elements
        n_neighbors_min = min([len(X_min), self.n_neighbors + 1])
        nn_min = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors_min,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
        nn_min.fit(X_min)
        ind_min = nn_min.kneighbors(X_min, return_distance=False)

        # implementation of sampling rules depending on the mode
        # the rules are not covering all cases

        if mode == 'high_complexity':
            samples = self.sampling_high_complexity(X_min,
                                                    labels,
                                                    ind_min,
                                                    n_to_sample)
        elif mode == 'low_complexity':
            samples = self.sampling_low_complexity(X_min,
                                                    labels,
                                                    ind_min,
                                                    n_to_sample)
        else:
            samples = self.sampling_otherwise(X_min,
                                                ind_min,
                                                n_to_sample)

        return samples

    def noise_removal(self, X, y, samples, nn_params):
        """
        Remove the noise from the genrerated samples.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            samples (np.array): the generated samples
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the updated samples
        """
        n_neighbors = min([len(X), self.n_neighbors + 1])

        # final noise removal by removing those minority samples generated
        # and not belonging to the lower approximation
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        ind_check = nnmt.kneighbors(samples, return_distance=False)

        num_maj_mask = \
            np.sum(y[ind_check[:, 1:]] == self.maj_label, axis=1) == 0

        return samples[num_maj_mask]

    def check_all_boundary(self, y):
        """
        Check if too many boundary samples identified.

        Args:
            y (np.array): the target labels
        """
        if np.sum(y == self.maj_label) <= np.sum(y == self.min_label):
            raise ValueError("too many majority samples identified as boundary")

    def check_labels(self, labels):
        """
        Check if not all minority samples are noise

        Args:
            labels (np.array): the minority labels
        """
        if len(np.unique(labels)) == 1 and labels[0] == 'NOI':
            raise ValueError("all minority samples identified as noise")

    def check_all_noise(self, y):
        """
        Check if there is only one class left after noise removal

        Args:
             y (np.array): the target labels
        """
        if len(np.unique(y)) == 1:
            raise ValueError("one class removed as noise")

    def all_steps(self, X, y, nn_params, n_to_sample):
        """
        Execute all steps of the algorithm.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            nn_params (dict): nearest neighbor parameters
            n_to_sample (int): the number of samples to generate
        """
        X_min, y = self.relabel_boundary(X, y, nn_params)

        self.check_all_boundary(y)

        labels = self.label_minority_samples(X, y, X_min, nn_params)

        self.check_labels(labels)

        mode = self.set_mode(X_min, labels)

        samples = self.generate_samples(X_min=X_min,
                                        nn_params=nn_params,
                                        n_to_sample=n_to_sample,
                                        labels=labels,
                                        mode=mode)

        samples = self.noise_removal(X, y, samples, nn_params)

        self.check_all_noise(y)

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # standardizing the data
        X_orig, y_orig = X, y

        scaler = StandardScaler()
        X = scaler.fit_transform(X)
        y = y.copy()

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        try:
            samples = self.all_steps(X, y, nn_params, n_to_sample)
        except ValueError as value_error:
            return self.return_copies(X_orig, y_orig, value_error.args[0])

        return (scaler.inverse_transform(np.vstack([X, samples])),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the SMOTE_PSO method.
"""

import numpy as np

from sklearn.svm import SVC
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import roc_auc_score

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['SMOTE_PSO']

class SMOTE_PSO(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_pso,
                        title = "PSO-based method for SVM classification on
                                    skewed data sets",
                        journal = "Neurocomputing",
                        volume = "228",
                        pages = "187 - 197",
                        year = "2017",
                        note = "Advanced Intelligent Computing: Theory and
                                    Applications",
                        issn = "0925-2312",
                        doi = "https://doi.org/10.1016/j.neucom.2016.10.041",
                        author = "Jair Cervantes and Farid Garcia-Lamont and
                                    Lisbeth Rodriguez and Asdrúbal López and
                                    José Ruiz Castilla and Adrian Trueba",
                        keywords = "Skew data sets, SVM, Hybrid algorithms"
                        }

    Notes:
        * I find the description of the technique a bit confusing, especially
            on the bounds of the search space of velocities and positions.
            Equations 15 and 16 specify the lower and upper bounds, the lower
            bound is in fact a vector while the upper bound is a distance.
            I tried to implement something meaningful.
        * I also find the setting of accelerating constant 2.0 strange, most
            of the time the velocity will be bounded due to this choice.
        * Also, training and predicting probabilities with a non-linear
            SVM as the evaluation function becomes fairly expensive when the
            number of training vectors reaches a couple of thousands. To
            reduce computational burden, minority and majority vectors far
            from the other class are removed to reduce the size of both
            classes to a maximum of 500 samples. Generally, this shouldn't
            really affect the results as the technique focuses on the samples
            near the class boundaries.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_memetic,
                  OverSampling.cat_uses_classifier,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 k=3,
                 *,
                 nn_params={},
                 eps=0.01,
                 n_pop=10,
                 w=1.0,
                 c1=2.0,
                 c2=2.0,
                 num_it=20,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            k (int): number of neighbors in nearest neighbors component, this
                        is also the multiplication factor of minority support
                        vectors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            eps (float): use to specify the initially generated support
                            vectors along minority-majority lines
            n_pop (int): size of population
            w (float): intertia constant
            c1 (float): acceleration constant of local optimum
            c2 (float): acceleration constant of population optimum
            num_it (int): number of iterations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(k, "k", 1)
        self.check_greater(eps, "eps", 0)
        self.check_greater_or_equal(n_pop, "n_pop", 1)
        self.check_greater_or_equal(w, "w", 0)
        self.check_greater_or_equal(c1, "c1", 0)
        self.check_greater_or_equal(c2, "c2", 0)
        self.check_greater_or_equal(num_it, "num_it", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.k = k # pylint: disable=invalid-name
        self.nn_params = nn_params
        self.params = {'eps': eps,
                       'n_pop': n_pop,
                       'w': w,
                       'c1': c1,
                       'c2': c2,
                       'num_it': num_it}
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        return cls.generate_parameter_combinations({'k': [3, 5, 7],
                                                    'eps': [0.01],
                                                    'n_pop': [5],
                                                    'w': [0.5, 1.0],
                                                    'c1': [1.0, 2.0],
                                                    'c2': [1.0, 2.0],
                                                    'num_it': [20]}, raw)

    # def remove_majority(self,
    #                     performance_th,
    #                     nn_params,
    #                     X_scaled, # pylint: disable=invalid-name
    #                     y):
    #     """
    #     Remove majority samples

    #     Args:
    #         performance_th (int): performance threshold
    #         nn_params (dict): nearest neighbor parameters
    #         X_scaled (np.array): scaled training vectors
    #         y (np.array): target labels

    #     Returns:
    #         np.array, np.array: the filtered training vectors and target labels
    #     """
    #     n_maj_to_remove = np.sum(y == self.maj_label) - performance_th

    #     if n_maj_to_remove <= 0:
    #         return X_scaled, y

    #     # if majority samples are to be removed
    #     nnmt= NearestNeighborsWithMetricTensor(n_neighbors=1,
    #                                             n_jobs=self.n_jobs,
    #                                             **nn_params)
    #     nnmt.fit(X_scaled[y == self.min_label])
    #     dist, _ = nnmt.kneighbors(X_scaled)

    #     dist = np.vstack([dist[0], np.arange(X_scaled.shape[0])]).T
    #     dist = dist[y == self.maj_label]
    #     dist = dist[dist[:, 0].argsort()[::-1]]

    #     to_remove = dist[:, 1][:n_maj_to_remove]

    #     # removing the samples
    #     X_scaled = np.delete(X_scaled, to_remove, axis=0)
    #     y = np.delete(y, to_remove)

    #     return X_scaled, y

    # def remove_minority(self,
    #                     performance_th,
    #                     nn_params,
    #                     X_scaled, # pylint: disable=invalid-name
    #                     y):
    #     """
    #     Remove minority samples

    #     Args:
    #         performance_th (int): performance threshold
    #         nn_params (dict): nearest neighbor parameters
    #         X_scaled (np.array): scaled training vectors
    #         y (np.array): target labels

    #     Returns:
    #         np.array, np.array: the filtered training vectors and target labels
    #     """

    #     n_min_to_remove = np.sum(y == self.min_label) - performance_th

    #     if n_min_to_remove <= 0:
    #         return X_scaled, y

    #     # if majority samples are to be removed
    #     nnmt = NearestNeighborsWithMetricTensor(n_neighbors=1,
    #                                             n_jobs=self.n_jobs,
    #                                             **nn_params)
    #     nnmt.fit(X_scaled[y == self.maj_label])

    #     dist, _ = nnmt.kneighbors(X_scaled)
    #     dist = np.vstack([dist[0], np.arange(X_scaled.shape[0])]).T
    #     dist = dist[y == self.min_label]
    #     dist = dist[dist[:, 0].argsort()[::-1]]

    #     to_remove = dist[:, 1][:n_min_to_remove]

    #     # removing the samples
    #     X_scaled = np.delete(X_scaled, to_remove, axis=0)
    #     y = np.delete(y, to_remove)

    #     return X_scaled, y

    def determine_support(self,
                            X_scaled, # pylint: disable=invalid-name
                            y):
        """
        Find the support.

        Args:
            X_scaled (np.array): all training vectors
            y (np.array): all target labels

        Returns:
            np.array, np.array: the minority and majority support vectors
        """
        # fitting SVM to extract initial support vectors
        svc = SVC(kernel='rbf', probability=True,
                  gamma='auto', random_state=self._random_state_init)
        svc.fit(X_scaled, y)

        # extracting the support vectors
        SV_min = np.array([i for i in svc.support_ if y[i] == self.min_label]) # pylint: disable=invalid-name
        SV_maj = np.array([i for i in svc.support_ if y[i] == self.maj_label]) # pylint: disable=invalid-name

        X_SV_min = X_scaled[SV_min] # pylint: disable=invalid-name
        X_SV_maj = X_scaled[SV_maj] # pylint: disable=invalid-name

        return X_SV_min, X_SV_maj

    def determine_indices(self, n_neighbors, nn_params, X_SV_maj, X_SV_min):
        """
        Determine the neighborhood relations.

        Args:
            n_neighbors (int): the number of neighbors to use
            nn_params (dict): nearest neighbors parameters
            X_SV_maj (np.array): majority vectors
            X_SV_min (np.array): minority vectors

        Returns:
            np.array: indices
        """
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_SV_maj)
        _, ind = nnmt.kneighbors(X_SV_min)

        return ind

    def initialize(self,
                    X_scaled, # pylint: disable=invalid-name
                    y,
                    nn_params):
        """
        Initialize the search

        Args:
            X_scaled (np.array): the scaled training vectors
            y (np.array): the target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array, np.array, np.array: the minority generators,
                                                    the search space,
                                                    the initial velocities,
                                                    and the search bounds
        """
        X_SV_min, X_SV_maj = self.determine_support(X_scaled, y) # pylint: disable=invalid-name

        # finding nearest majority support vectors
        n_neighbors = np.min([X_SV_maj.shape[0], self.k])
        ind = self.determine_indices(n_neighbors, nn_params, X_SV_maj, X_SV_min)

        min_vector = X_SV_min[np.repeat(np.arange(len(X_SV_min)), n_neighbors)]
        maj_vector = X_SV_maj[ind.flatten()]
        upper_bound = X_SV_maj[ind[np.repeat(np.arange(len(X_SV_min)), n_neighbors), 0]]
        init_velocity = self.params['eps'] * (maj_vector - min_vector)
        X_min_gen = min_vector + init_velocity # pylint: disable=invalid-name

        search_bound = np.linalg.norm(min_vector - upper_bound, axis=1)

        search_space = np.hstack([min_vector, maj_vector]).reshape(min_vector.shape[0],
                                                                    2,
                                                                    min_vector.shape[1])

        return X_min_gen, search_space, search_bound, init_velocity

    def evaluate(self, X_train, y_train, X_test, y_test):
        """
        Trains support vector classifier and evaluates it

        Args:
            X_train (np.array): training vectors
            y_train (np.array): target labels
            X_test (np.array): test vectors
            y_test (np.array): test labels
        """
        svc = SVC(kernel='rbf', probability=True,
                  gamma='auto', random_state=self._random_state_init)

        svc.fit(X_train, y_train)
        class_idx = np.where(svc.classes_ == self.min_label)[0][0]
        y_pred = svc.predict_proba(X_test)[:, class_idx]
        return roc_auc_score(y_test, y_pred)

    def evaluate_particle(self,
                            X_scaled, # pylint: disable=invalid-name
                            y,
                            part):
        """
        Evaluate a particle.

        Args:
            X_scaled (np.array): the training vectors
            y (np.array): the target labels
            part (np.array): a particle
        """
        X_extended = np.vstack([X_scaled, part]) # pylint: disable=invalid-name
        y_extended = np.hstack([y, np.repeat(self.min_label, len(part))])
        return self.evaluate(X_extended, y_extended, X_scaled, y)

    def update_velocities(self, *, particle_swarm, velocities,
                            search_bound, local_best, global_best):
        """
        Update the velocities.

        Args:
            particle_swarm (np.array): the particles
            velocities (np.array): the velocities
            search_bound (np.array): the search bounds
            local_best (np.array): the local best results
            global_best (np.array): the global best result

        Returns:
            list: the updated velocities
        """
        # update velocities
        #for idx, part in enumerate(particle_swarm):
        #    term_0 = self.params['w'] * velocities[idx]
        #    rand = self.random_state.random_sample(2)
        #    term_1 = self.params['c1'] * rand[0] * (local_best[idx] - part)
        #    term_2 = self.params['c2'] * rand[1] * (global_best - part)
        #
        #    velocities[idx] = term_0 + term_1 + term_2
        #    #velocities[idx] = term_0

        term_0 = self.params['w'] * velocities
        rand = self.random_state.random_sample((2, velocities.shape[0]))
        term_1 = self.params['c1'] * (local_best - particle_swarm) \
                                                        * rand[0][:, None, None]
        term_2 = self.params['c2'] * (global_best - particle_swarm) \
                                                        * rand[1][:, None, None]

        velocities = term_0 + term_1 + term_2

        velocity_norms = np.linalg.norm(velocities, axis=2)

        mask = (velocity_norms > search_bound[None, :] / 2.0)

        search_bounds = np.tile(search_bound[:, None], (velocity_norms.shape[0],)).T

        multiplier = search_bounds[mask] / velocity_norms[mask] / 2.0

        velocities[mask] = velocities[mask] * multiplier[:, None]

        # bound velocities according to search space constraints
        #for vel in velocities:
        #    for idx, vel_idx in enumerate(vel):
        #        v_i_norm = np.linalg.norm(vel_idx)
        #        if v_i_norm > search_bound[idx]/2.0:
        #            vel[idx] = vel_idx / v_i_norm * search_bound[idx] / 2.0

        return velocities

    def update_positions(self, particle_swarm, velocities, search_space, search_bound):
        """
        Update the positions.

        Args:
            particle_swarm (np.array): the particles
            velocities (np.array): the velocities
            search_space (np.array): the search space
            search_bound (np.array): the search bounds

        Returns:
            np.array: the updated particle swarm
        """
        # update positions
        particle_swarm = particle_swarm + velocities

        trans_vectors = particle_swarm - search_space[:, 0][None, :, :]
        trans_norm = np.linalg.norm(trans_vectors, axis=2)

        mask = trans_norm > search_bound[None, :]

        search_space_0 = np.tile(search_space[:, 0][None, :, :], (particle_swarm.shape[0], 1, 1))

        search_bounds = np.tile(search_bound, (particle_swarm.shape[0], 1))

        multiplier = search_bounds[mask] / trans_norm[mask]

        normed_trans = trans_vectors[mask] * multiplier[:, None]

        particle_swarm[mask] = search_space_0[mask] + normed_trans

        #for idx, part in enumerate(particle_swarm):
        #    particle_swarm[idx] = particle_swarm[idx] + velocities[idx]

        # bound positions according to search space constraints
        #for part in particle_swarm:
        #    for idx, p_idx in enumerate(part):
        #        search_idx = search_space[idx]
        #
        #        trans_vector = p_idx - search_idx[0]
        #        trans_norm = np.linalg.norm(trans_vector)
        #
        #        if trans_norm > search_idx[2]:
        #            normed_trans = trans_vector / trans_norm
        #            part[idx] = search_idx[0] + normed_trans * search_idx[2]

        return particle_swarm

    def update_scores(self, *, X_scaled, y, particle_swarm,
                local_best_scores, local_best, global_best_score, global_best):
        """
        Evaluate and update the scores.

        Args:
            X_scaled (np.array): the training vectors
            y (np.array): the target labels
            particle_swarm (np.array): the particle swarm
            local_best_scores (np.array): the local best scores
            local_best (np.array): the local best particles
            global_best_score (float): the global best score
            global_best (obj): the globally best particle

        Returns:
            np.array, np.array, float, obj: the updated local best scores,
                        local best, global best score and global best
                        particle
        """

        scores = np.array([self.evaluate_particle(X_scaled, y, p)
                      for p in particle_swarm])

        local_mask = scores > local_best_scores
        local_best_scores[local_mask] = scores[local_mask]
        local_best[local_mask] = particle_swarm[local_mask]

        max_idx = np.argmax(scores)
        if scores[max_idx] > global_best_score:
            global_best_score = scores[max_idx]
            global_best = particle_swarm[max_idx]

        # update best scores
        #for idx, score in enumerate(scores):
        #    if score > local_best_scores[idx]:
        #        local_best_scores[idx] = score
        #        local_best[idx] = particle_swarm[idx]
        #    if score > global_best_score:
        #        global_best_score = score
        #        global_best = particle_swarm[idx]

        return local_best_scores, local_best, global_best_score, global_best

    def search(self,
                X_scaled, # pylint: disable=invalid-name
                y,
                nn_params):
        """
        The optimization search.

        Args:
            X_scaled (np.array): the scaled training vectors
            y (np.array): the target label
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the generated samples
        """
        X_min_gen, search_space, search_bound, init_velocity = \
                    self.initialize(X_scaled, y, nn_params) # pylint: disable=invalid-name

        # initializing the particle swarm and the particle and population level
        # memory
        particle_swarm = np.array([X_min_gen.copy()] * self.params['n_pop'])
        velocities = np.array([init_velocity.copy()] * self.params['n_pop'])
        local_best = np.array([X_min_gen.copy()] * self.params['n_pop'])
        local_best_scores = np.array([0.0]*self.params['n_pop'])
        global_best = X_min_gen.copy()
        global_best_score = 0.0

        for iteration in range(self.params['num_it']):
            _logger.info("%s: Iteration %d", self.__class__.__name__, iteration)
            # evaluate population
            local_best_scores, local_best, global_best_score, global_best = \
                self.update_scores(X_scaled=X_scaled,
                                    y=y,
                                    particle_swarm=particle_swarm,
                                    local_best_scores=local_best_scores,
                                    local_best=local_best,
                                    global_best_score=global_best_score,
                                    global_best=global_best)

            velocities = self.update_velocities(particle_swarm=particle_swarm,
                                                velocities=velocities,
                                                search_bound=search_bound,
                                                local_best=local_best,
                                                global_best=global_best)

            particle_swarm = self.update_positions(particle_swarm,
                                                    velocities,
                                                    search_space,
                                                    search_bound)
        return global_best

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        if np.sum(y == self.min_label) >= np.sum(y == self.maj_label):
            return self.return_copies(X, y, "Sampling is not needed.")

        # saving original dataset
        X_orig = X
        y_orig = y

        # scaling the records
        #mms = MinMaxScaler()
        mms = StandardScaler()
        X_scaled = mms.fit_transform(X) # pylint: disable=invalid-name

        # removing majority and minority samples far from the training data if
        # needed to increase performance
        #performance_th = 500

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X_scaled, y)

        #X_scaled, y = self.remove_majority(performance_th, nn_params, X_scaled, y)

        #X_scaled, y = self.remove_minority(performance_th, nn_params, X_scaled, y)

        new_items = [np.zeros(shape=(0, X.shape[1]))]

        while np.sum(y == self.min_label) < np.sum(y == self.maj_label):
            global_best = self.search(X_scaled, y, nn_params)
            new_items.append(global_best)
            X_scaled = np.vstack([X_scaled, global_best]) # pylint: disable=invalid-name
            y = np.hstack([y, np.repeat(self.min_label, len(global_best))])

        X_new = np.vstack(new_items)
        X_ret = np.vstack([X_orig, mms.inverse_transform(X_new)]) # pylint: disable=invalid-name
        y_ret = np.hstack([y_orig, np.repeat(self.min_label, len(X_new))])

        return (X_ret, y_ret)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'k': self.k,
                'nn_params': self.nn_params,
                'eps': self.params['eps'],
                'n_pop': self.params['n_pop'],
                'w': self.params['w'],
                'c1': self.params['c1'],
                'c2': self.params['c2'],
                'num_it': self.params['num_it'],
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the Assembled_SMOTE.
"""

import numpy as np

from smote_variants import Assembled_SMOTE

def test_specific():
    """
    Oversampler specific testing
    """

    obj = Assembled_SMOTE()

    vectors = [np.array([[1.0, 2.0]]), np.array([[2.0, 3.0]])]

    samples = obj.generate_samples_in_clusters(vectors=vectors,
                                                n_to_sample=4,
                                                nn_params={})
    assert len(samples) == 4



"""
Testing the NDO_sampling.
"""
import numpy as np

from smote_variants import NDO_sampling

from smote_variants.datasets import load_normal

def test_specific():
    """
    Oversampler specific testing
    """
    obj = NDO_sampling(T=20.0)

    dataset = load_normal()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert X_samp.shape[0] > 0

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9]])
    y = np.array([0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    alpha = obj.determine_alpha(y = np.array([0, 0, 1, 1]),
                                ind = np.array([[2, 0],
                                                [3, 1]]),
                                dist = np.array([[0.0, 0.0],
                                                [0.0, 0.0]]))
    assert alpha == 0.0

    alpha = obj.determine_alpha(y = np.array([0, 0, 1, 1]),
                                ind = np.array([[2, 3],
                                                [3, 2]]),
                                dist = np.array([[0.0, 0.0],
                                                [0.0, 0.0]]))
    assert alpha == np.inf

    alpha = obj.determine_alpha(y = np.array([0, 0, 1, 1]),
                                ind = np.array([[2, 3],
                                                [3, 1]]),
                                dist = np.array([[0.0, 0.0],
                                                [0.0, 0.0]]))
    assert alpha == np.inf



"""
This module implements the SL_graph_SMOTE method.
"""

import numpy as np

from scipy.stats import skew

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from ._safe_level_smote import Safe_Level_SMOTE
from ._borderline_smote import Borderline_SMOTE1

from .._logger import logger
_logger= logger

__all__= ['SL_graph_SMOTE']

class SL_graph_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @inproceedings{sl_graph_smote,
                    author = {Bunkhumpornpat,
                        Chumpol and Subpaiboonkit, Sitthichoke},
                    booktitle= {13th International Symposium on Communications
                                and Information Technologies},
                    year = {2013},
                    month = {09},
                    pages = {570-575},
                    title = {Safe level graph for synthetic minority
                                over-sampling techniques},
                    isbn = {978-1-4673-5578-0}
                    }
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_borderline,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                    to sample e.g. 1.0 means that after
                                    sampling the number of minority samples
                                    will be equal to the number of majority
                                    samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # Fitting nearest neighbors model
        n_neighbors = min([len(X), self.n_neighbors])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        indices = nnmt.kneighbors(X[y == self.min_label], return_distance=False)

        # Computing safe level values
        safe_level_values = np.sum(y[indices], axis=1)

        # Computing skewness
        skewness = skew(safe_level_values)

        if skewness < 0:
            # left skewed
            sampler = Safe_Level_SMOTE(proportion=self.proportion,
                                        n_neighbors=self.n_neighbors,
                                        nn_params=nn_params,
                                        ss_params=self.ss_params,
                                        n_jobs=self.n_jobs,
                                        random_state=self._random_state_init)
        else:
            # right skewed
            sampler = Borderline_SMOTE1(proportion=self.proportion,
                                        n_neighbors=self.n_neighbors,
                                        nn_params=nn_params,
                                        ss_params=self.ss_params,
                                        n_jobs=self.n_jobs,
                                        random_state=self._random_state_init)

        return sampler.sample(X, y)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements some basic functionalities.
"""

import itertools
import importlib
import json
import pickle

from json import JSONDecodeError

import numpy as np

from .._logger import logger
_logger = logger

__all__= ['cov',
          'mode',
          'fix_density',
          'safe_divide',
          'equal_dicts',
          'coalesce',
          'coalesce_dict',
          'instantiate_obj',
          'load_dict',
          'dump_dict',
          'check_if_damaged',
          'StatisticsMixin',
          'RandomStateMixin',
          'ParametersMixin']

def cov(array, rowvar=True):
    """
    Wrapper over the numpy covariance function to handle
    the 1 feature case better.

    Args:
        array (np.array): to compute the covariance for
        rowvar (bool): whether the rows represent the observations
    """

    if not rowvar:
        if array.shape[1] > 1:
            return np.cov(array, rowvar=rowvar)
        return np.array([[np.cov(array, rowvar=rowvar)]])

    if array.shape[0] > 1:
        return np.cov(array, rowvar=rowvar)
    return np.array([[np.cov(array, rowvar=rowvar)]])

def fix_density(density):
    """
    Create a valid density distribution.
    """
    density = np.nan_to_num(density, nan=0.0, posinf=0.0, neginf=0.0)

    if np.sum(density) == 0.0:
        density = np.repeat(1.0, len(density))

    return density / np.sum(density)

def safe_divide(numerator, denominator, default=np.inf):
    """
    Safe division.

    Args:
        numerator (numeric): the numerator
        denominator (numeric): the denominator
        default (numeric): the value to return in case of zero division
    """
    if denominator == 0:
        return default
    return numerator/denominator

def equal_dicts(dict_0, dict_1):
    """
    Compare literal dictionaries.

    Args:
        dict_0 (dict): dictionary 0
        dict_1 (dict): dictionary 1

    Returns:
        bool: whether the dicts are equal
    """
    if len(dict_0) != len(dict_1):
        return False

    for key, val in dict_0.items():
        if isinstance(val, dict):
            if not equal_dicts(val, dict_1[key]):
                return False
        else:
            if val != dict_1[key]:
                return False

    return True

def coalesce(value_0, value_1):
    """
    The coalesce functionality.

    Args:
        value_0 (obj): first value
        value_1 (obj): second value

    Returns:
        obj: value_1 if value_0 is None otherwise value_0
    """
    if value_0 is None:
        return value_1
    return value_0

def coalesce_dict(dict_0, dict_1):
    """
    Coalesce dicts key by key

    Args:
        dict_0 (dict): the first dict
        dict_1 (dict): the second dict

    Returns:
        dict: the coalesced dict
    """
    if dict_0 is None:
        return dict_1

    result = {**dict_1}
    for key, item in dict_0.items():
        result[key] = item

    return result


def instantiate_obj(description):
    """
    Instantiates an object from a description

    Args:
        description (tuple): (module_name, class_name, params_dict)

    Returns:
        obj: the instantiated object
    """

    module = importlib.import_module(description[0])
    class_ = getattr(module, description[1])
    return class_(**description[2])

def mode(data):
    """
    Returns the mode of the data

    Args:
        data (np.array/list): the data to compute the mode of

    Returns:
        float: the mode of the data
    """
    values, counts = np.unique(data, return_counts=True)
    return values[np.where(counts == max(counts))[0][0]]

def load_dict(filename,
                serialization=None,
                array_to_list_map=None):
    """
    Load the contents of a file

    Args:
        filename (str): the filename to load
        serialization (str): 'json'/'pickle'/None
        array_to_list_map (list/None): list of keys to convert

    Returns:
        obj: the loaded object
    """
    if array_to_list_map is None:
        array_to_list_map = []

    if serialization == 'json':
        with open(filename, 'rt', encoding='UTF-8') as file:
            obj = json.load(file)
            for key in array_to_list_map:
                obj[key] = np.array(obj[key])
            return obj
    elif serialization == 'pickle':
        with open(filename, 'rb') as file:
            return pickle.load(file)
    else:
        raise ValueError(f"serialization {serialization} is not supported")

def dump_dict(obj,
                filename,
                serialization=None,
                array_to_list_map=None):
    """
    Serializes an object.

    Args:
        obj (obj): object to serialize
        filename (str): the filename
        serialization (str): 'json'/pickle'/None
        array_to_list_map (list/None): list of keys to convert
    """
    if array_to_list_map is None:
        array_to_list_map = []

    if serialization == 'json':
        for key in array_to_list_map:
            obj[key] = obj[key].astype(float).tolist()
        with open(filename, 'wt', encoding='UTF-8') as file:
            json.dump(obj, file)
    elif serialization == 'pickle':
        with open(filename, 'wb') as file:
            pickle.dump(obj, file)
    else:
        raise ValueError(f"serialization {serialization} is not supported")

def check_if_damaged(filename, serialization="json"):
    """
    Check if a file is available and not damaged

    Args:
        filename (str): the filename
        serialization (str): the type of serialization ('json'/'pickle')

    Returns:
        bool: True if the file is damaged or not available
    """
    damaged = False

    try:
        load_dict(filename, serialization)
    except FileNotFoundError:
        damaged = True
    except JSONDecodeError:
        damaged = True

    return damaged

class StatisticsMixin:
    """
    Mixin to compute class statistics and determine minority/majority labels
    """

    def class_label_statistics(self, y):
        """
        determines class sizes and minority and majority labels
        Args:
            X (np.array): features
            y (np.array): target labels
        """
        unique, counts = np.unique(y, return_counts=True)
        self.class_stats = dict(zip(unique, counts))
        self.min_label = unique[0] if counts[0] < counts[1] else unique[1]
        self.maj_label = unique[1] if counts[0] < counts[1] else unique[0]

    def get_params(self, deep=False):
        """
        Return the parameters

        Returns:
            dict: the parameters of the mixin
        """
        _ = deep
        return {}

    #def check_enough_min_samples_for_sampling(self, threshold=2):
    #    """
    #    Checks if there are enough samples for oversampling, if not,
    #    logs it and returns false
    #
    #    Args:
    #        threshold (int): the minimum number of minority samples
    #
    #    Returns:
    #        str/None: string massage about the problem
    #    """
    #    msg = None
    #    if self.class_stats[self.min_label] < threshold:
    #        msg = "The number of minority samples "\
    #                f"({self.class_stats[self.min_label]}) is not enough "\
    #                "for sampling"
    #    return msg

    #def check_n_to_sample(self, n_to_sample):
    #    """
    #    Check if there is need for sampling.
    #
    #    Args:
    #        n_to_sample (int): number of samples to generate
    #
    #    Returns:
    #        str/None: string message about the problem
    #    """
    #
    #    msg = None
    #    if n_to_sample == 0:
    #        msg = "There is no need for oversampling"
    #    return msg

class RandomStateMixin:
    """
    Mixin to set random state
    """
    def __init__(self, random_state):
        """
        Constructor of the mixin

        Args:
            random_state (int/np.random.RandomState/None): the random state
                                                                initializer
        """
        self.set_random_state(random_state)

    def set_random_state(self, random_state):
        """
        sets the random_state member of the object

        Args:
            random_state (int/np.random.RandomState/None): the random state
                                                                initializer
        """

        self._random_state_init = random_state

        if random_state is None:
            self.random_state = np.random
        elif isinstance(random_state, int):
            self.random_state = np.random.RandomState(random_state)
        elif isinstance(random_state, np.random.RandomState):
            self.random_state = random_state
        elif random_state is np.random:
            self.random_state = random_state
        else:
            raise ValueError(
                "random state cannot be initialized by " + str(random_state))

    def get_params(self, deep=False):
        """
        Returns the parameters of the object.

        Args:
            deep (bool): deep parameters

        Returns:
            dict: the parameter dictionary
        """
        _ = deep # disabling pylint reporting
        return {'random_state': self._random_state_init}

class ParametersMixin:
    """
    Mixin to check if parameters come from a valid range
    """

    def check_in_range(self, value, name, rng):
        """
        Check if parameter is in range
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            rng (list-like(2)): the lower and upper bound of a range
        Throws:
            ValueError
        """
        if value < rng[0] or value > rng[1]:
            msg = ("Value for parameter %s outside the range [%f,%f] not"
                 " allowed: %f")
            msg = msg % (name, rng[0], rng[1], value)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_out_range(self, value, name, rng):
        """
        Check if parameter is outside of range
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            rng (list-like(2)): the lower and upper bound of a range
        Throws:
            ValueError
        """
        if rng[0] <= value <= rng[1]:
            msg = "Value for parameter %s in the range [%f,%f] not allowed: %f"
            msg = msg % (name, rng[0], rng[1], value)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_less_or_equal(self, value, name, val):
        """
        Check if parameter is less than or equal to value
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            val (numeric): value to compare to
        Throws:
            ValueError
        """
        if value > val:
            msg = "Value for parameter %s greater than %f not allowed: %f > %f"
            msg = msg % (name, val, value, val)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_less_or_equal_par(self, val_x, name_x, val_y, name_y):
        """
        Check if parameter is less than or equal to another parameter
        Args:
            val_x (numeric): the parameter value
            name_x (str): the parameter name
            val_y (numeric): the other parameter value
            name_y (str): the other parameter name
        Throws:
            ValueError
        """
        if val_x > val_y:
            msg = ("Value for parameter %s greater than parameter %s not"
                 " allowed: %f > %f")
            msg = msg % (name_x, name_y, val_x, val_y)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_less(self, value, name, val):
        """
        Check if parameter is less than value
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            val (numeric): value to compare to
        Throws:
            ValueError
        """
        if value >= val:
            msg = ("Value for parameter %s greater than or equal to %f"
                 " not allowed: %f >= %f")
            msg = msg % (name, val, value, val)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_less_par(self, val_x, name_x, val_y, name_y):
        """
        Check if parameter is less than another parameter
        Args:
            val_x (numeric): the parameter value
            name_x (str): the parameter name
            val_y (numeric): the other parameter value
            name_y (str): the other parameter name
        Throws:
            ValueError
        """
        if val_x >= val_y:
            msg = ("Value for parameter %s greater than or equal to parameter"
                 " %s not allowed: %f >= %f")
            msg = msg % (name_x, name_y, val_x, val_y)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_greater_or_equal(self, value, name, val):
        """
        Check if parameter is greater than or equal to value
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            val (numeric): value to compare to
        Throws:
            ValueError
        """
        if value < val:
            msg = "Value for parameter %s less than %f is not allowed: %f < %f"
            msg = msg % (name, val, value, val)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_greater_or_equal_par(self, val_x, name_x, val_y, name_y):
        """
        Check if parameter is less than or equal to another parameter
        Args:
            val_x (numeric): the parameter value
            name_x (str): the parameter name
            val_y (numeric): the other parameter value
            name_y (str): the other parameter name
        Throws:
            ValueError
        """
        if val_x < val_y:
            msg = ("Value for parameter %s less than parameter %s is not"
                 " allowed: %f < %f")
            msg = msg % (name_x, name_y, val_x, val_y)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_greater(self, value, name, val):
        """
        Check if parameter is greater than value
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            val (numeric): value to compare to
        Throws:
            ValueError
        """
        if value <= val:
            msg = ("Value for parameter %s less than or equal to %f not allowed"
                 " %f < %f")
            msg = msg % (name, val, value, val)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_greater_par(self, val_x, name_x, val_y, name_y):
        """
        Check if parameter is greater than or equal to another parameter
        Args:
            val_x (numeric): the parameter value
            name_x (str): the parameter name
            val_y (numeric): the other parameter value
            name_y (str): the other parameter name
        Throws:
            ValueError
        """
        if val_x <= val_y:
            msg = ("Value for parameter %s less than or equal to parameter %s"
                 " not allowed: %f <= %f")
            msg = msg % (name_x, name_y, val_x, val_y)

            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_equal(self, value, name, val):
        """
        Check if parameter is equal to value
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            val (numeric): value to compare to
        Throws:
            ValueError
        """
        if value == val:
            msg = ("Value for parameter %s equal to parameter %f is not allowed:"
                 " %f == %f")
            msg = msg % (name, val, value, val)
            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_equal_par(self, val_x, name_x, val_y, name_y):
        """
        Check if parameter is equal to another parameter
        Args:
            val_x (numeric): the parameter value
            name_x (str): the parameter name
            val_y (numeric): the other parameter value
            name_y (str): the other parameter name
        Throws:
            ValueError
        """
        if val_x == val_y:
            msg = ("Value for parameter %s equal to parameter %s is not "
                 "allowed: %f == %f")
            msg = msg % (name_x, name_y, val_x, val_y)
            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_isin(self, value, name, list_):
        """
        Check if parameter is in list
        Args:
            value (numeric): the parameter value
            name (str): the parameter name
            list_ (list): list to check if parameter is in it
        Throws:
            ValueError
        """
        if value not in list_:
            msg = "Value for parameter %s not in list %s is not allowed: %s"
            msg = msg % (name, str(list_), str(value))
            raise ValueError(self.__class__.__name__ + ": " + msg)

    def check_n_jobs(self, value, name='n_jobs'):
        """
        Check n_jobs parameter
        Args:
            value (int/None): number of jobs
            name (str): the parameter name
        Throws:
            ValueError
        """
        if ((value is None)
                or (value is not None and isinstance(value, int) and value <= 0)):
            msg = "Value for parameter %s is not allowed: %s"
            msg = msg % (name, str(value))
            raise ValueError(self.__class__.__name__ + ": " + msg)

    @classmethod
    def generate_parameter_combinations(cls, dictionary, raw):
        """
        Generates reasonable parameter combinations
        Args:
            dictionary (dict): dictionary of parameter ranges
            num (int): maximum number of combinations to generate
        """
        if raw:
            return dictionary
        keys = sorted(list(dictionary.keys()))
        values = [dictionary[k] for k in keys]
        combinations = [dict(zip(keys, p))
                        for p in list(itertools.product(*values))]
        return combinations



"""
Tests the operation of the MLPWrapper
"""

import logging

from sklearn import datasets

import smote_variants as sv

logger = logging.getLogger("smote_variants")
logger.setLevel(logging.ERROR)

def test_mlp_wrapper():
    """
    Testing the MLP wrapper.
    """
    dataset = datasets.load_wine()
    classifier = sv.classifiers.MLPClassifierWrapper()
    classifier.fit(dataset['data'], dataset['target'])

    assert classifier is not None

    pred = classifier.predict(dataset['data'])

    assert len(pred) == len(dataset['data'])

    pred_proba = classifier.predict_proba(dataset['data'])

    assert len(pred_proba) == len(dataset['data'])

    params = classifier.get_params()

    assert len(params) > 0

    clone = classifier.copy()

    assert sv.base.equal_dicts(params, clone.get_params())



"""
This module implements the AHC method
"""
import warnings

import numpy as np

from sklearn.cluster import AgglomerativeClustering, KMeans

from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['AHC']

class AHC(OverSampling):
    """
    References:
        * BibTex::

            @article{AHC,
                    title = "Learning from imbalanced data in surveillance
                             of nosocomial infection",
                    journal = "Artificial Intelligence in Medicine",
                    volume = "37",
                    number = "1",
                    pages = "7 - 18",
                    year = "2006",
                    note = "Intelligent Data Analysis in Medicine",
                    issn = "0933-3657",
                    doi = "https://doi.org/10.1016/j.artmed.2005.03.002",
                    url = {http://www.sciencedirect.com/science/article/
                            pii/S0933365705000850},
                    author = "Gilles Cohen and Mélanie Hilario and Hugo Sax
                                and Stéphane Hugonnet and Antoine Geissbuhler",
                    keywords = "Nosocomial infection, Machine learning,
                                    Support vector machines, Data imbalance"
                    }
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_uses_clustering,
                  OverSampling.cat_application]

    def __init__(self,
                 *,
                 strategy='min',
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            strategy (str): which class to sample (min/maj/minmaj)
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_isin(strategy, 'strategy', ['min', 'maj', 'minmaj'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.strategy = strategy
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'strategy': ['min', 'maj', 'minmaj']}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sample_majority(self, X, n_clusters):
        """
        Sample the majority class

        Args:
            X (np.ndarray): majority samples
            n_clusters (int): number of clusters to find

        Returns:
            np.ndarray: downsampled vectors
        """
        kmeans = KMeans(n_clusters=n_clusters,
                        random_state=self._random_state_init)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            kmeans.fit(X)
        return kmeans.cluster_centers_

    def sample_minority(self, X):
        """
        Sampling the minority class

        Args:
            X (np.ndarray): minority samples

        Returns:
            np.ndarray: the oversampled set of vectors
        """
        aggc = AgglomerativeClustering(n_clusters=1)
        aggc.fit(X)
        n_samples = len(X)

        clus_cent = [None]*len(aggc.children_)
        weights = [None]*len(aggc.children_)

        def cluster_centers(children, idx, clus_cent, weights):
            """
            Extract cluster centers

            Args:
                children (np.array): indices of children
                idx (int): index to process
                clus_cent (np.array): cluster centers
                weights (np.array): cluster weights

            Returns:
                int, float: new cluster center, new weight
            """
            if idx < n_samples:
                return X[idx], 1.0

            if clus_cent[idx - n_samples] is None:
                a_cent, w_a = cluster_centers(
                    children, children[idx - n_samples][0], clus_cent, weights)
                b_cent, w_b = cluster_centers(
                    children, children[idx - n_samples][1], clus_cent, weights)
                clus_cent[idx - n_samples] = (w_a*a_cent + w_b*b_cent)/(w_a + w_b)
                weights[idx - n_samples] = w_a + w_b

            return clus_cent[idx - n_samples], weights[idx - n_samples]

        cluster_centers(aggc.children_, aggc.children_[-1][-1] + 1, clus_cent, weights)

        return np.vstack(clus_cent)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        # extracting minority samples
        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        if self.strategy == 'maj':
            X_maj_resampled = self.sample_majority(X_maj, len(X_min)) # pylint: disable=invalid-name
            return (np.vstack([X_min, X_maj_resampled]),
                    np.hstack([np.repeat(self.min_label, len(X_min)),
                               np.repeat(self.maj_label,
                                         len(X_maj_resampled))]))
        if self.strategy == 'min':
            X_min_resampled = self.sample_minority(X_min) # pylint: disable=invalid-name
            return (np.vstack([X_min_resampled, X_min, X_maj]),
                    np.hstack([np.repeat(self.min_label,
                                         (len(X_min_resampled) + len(X_min))),
                               np.repeat(self.maj_label, len(X_maj))]))
        # the "minmaj" strategy case
        X_min_resampled = self.sample_minority(X_min) # pylint: disable=invalid-name
        n_maj_sample = min([len(X_maj), len(X_min_resampled) + len(X_min)])
        X_maj_resampled = self.sample_majority(X_maj, n_maj_sample) # pylint: disable=invalid-name
        return (np.vstack([X_min_resampled, X_min, X_maj_resampled]),
                np.hstack([np.repeat(self.min_label,
                                        (len(X_min_resampled) + len(X_min))),
                            np.repeat(self.maj_label,
                                        len(X_maj_resampled))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'strategy': self.strategy,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the SNSMOTE method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import (NearestNeighborsWithMetricTensor,
                            distances_mahalanobis)
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['SN_SMOTE']

class SN_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @Article{sn_smote,
                        author="Garc{\'i}a, V.
                        and S{\'a}nchez, J. S.
                        and Mart{\'i}n-F{\'e}lez, R.
                        and Mollineda, R. A.",
                        title="Surrounding neighborhood-based SMOTE for
                                learning from imbalanced data sets",
                        journal="Progress in Artificial Intelligence",
                        year="2012",
                        month="Dec",
                        day="01",
                        volume="1",
                        number="4",
                        pages="347--362",
                        issn="2192-6360",
                        doi="10.1007/s13748-012-0027-5",
                        url="https://doi.org/10.1007/s13748-012-0027-5"
                        }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                    to sample e.g. 1.0 means that after
                                    sampling the number of minority samples
                                    will be equal to the number of majority
                                    samples
            n_neighbors (float): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_centroid_neighborhood(self, X_min, nn_params, ind):
        """
        Generate the centroid neighborhood structure.

        Args:
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbor parameters
            ind (np.array): the neighborhood structure

        """
        n_neighbors = ind.shape[1]

        # determining k nearest centroid neighbors
        ncn = np.zeros(shape=(X_min.shape[0], self.n_neighbors)).astype(int)

        metric_tensor = coalesce(nn_params.get('metric_tensor', None),
                                 np.eye(X_min.shape[1]))

        ncn[:, 0] = ind[:, 1]
        n_cents = np.ones(ncn.shape[0]).astype(int)
        centroids = X_min[ncn[:, 0]]
        cent_dist = distances_mahalanobis(X=X_min,
                                            Y=centroids,
                                            tensor=metric_tensor)

        mask = np.repeat(True, X_min.shape[0])
        jdx = 2
        while np.any(mask):
            centroids_tmp = ((centroids[mask] + X_min[ind[mask, jdx]]).T / (n_cents[mask] + 1)).T
            new_cent_dist = distances_mahalanobis(X=X_min[mask],
                                                    Y=centroids_tmp,
                                                    tensor=metric_tensor)

            improved = new_cent_dist < cent_dist[mask]
            centroids[mask][improved] = centroids[mask][improved] + X_min[ind[mask, jdx][improved]]
            ncn[mask][improved, n_cents[mask][improved]] = ind[mask][improved, jdx]
            n_cents[mask][improved] = n_cents[mask][improved] + 1
            cent_dist[mask][improved] = new_cent_dist[improved]

            mask = (n_cents < self.n_neighbors)

            jdx = jdx + 1
            if jdx == n_neighbors:
                break

        return ncn, n_cents

    def generate_samples(self, X_min, n_to_sample, ncn, ncn_nums):
        """
        Generate the samples.

        Args:
            X_min (np.array): minortity samples
            n_to_sample (int): the number of samples to generate
            ncn (np.array): the nearest centroid neighborhood
            ncn_nums (np.array): the sizes of the neighborhoods

        Returns:
            np.array: the generated samples
        """
        samples = np.zeros(shape=(0, X_min.shape[1]))

        base_indices = self.random_state.choice(X_min.shape[0], n_to_sample)
        base_unique, base_count = np.unique(base_indices, return_counts=True)

        for idx, base_idx in enumerate(base_unique):
            X_vertices = X_min[ncn[base_idx, :ncn_nums[base_idx]]]
            indices = np.array([np.hstack([np.array([0]),
                                np.arange(ncn_nums[base_idx])])])

            #n_dim_orig = self.n_dim
            #self.n_dim = np.min([ncn_nums[base_idx] + 1, n_dim_orig])

            samples_tmp = self.sample_simplex(X=X_min[[base_idx]],
                                                indices=indices,
                                                n_to_sample=base_count[idx],
                                                X_vertices=X_vertices)
            #self.n_dim = n_dim_orig

            samples = np.vstack([samples, samples_tmp])

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # the search for the k nearest centroid neighbors is limited for the
        # nearest 10*n_neighbors neighbors

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        n_neighbors = min([self.n_neighbors*10, len(X_min)])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        ncn, ncn_nums = self.generate_centroid_neighborhood(X_min, nn_params, ind)

        # extracting nearest centroid neighbors
        #for idx, x_min in enumerate(X_min):
        #    # the first NCN neighbor is the first neighbor
        #    ncn[idx, 0] = ind[idx][1]
        #
        #    # iterating through all neighbors and finding the one with smaller
        #    # centroid distance to X_min[i] than the previous set of neighbors
        #    n_cent = 1
        #    centroid = X_min[ncn[idx, 0]]
        #
        #    cent_dist = np.sqrt(np.dot(np.dot((centroid - x_min), metric_tensor),
        #                                (centroid - x_min)))
        #    jdx = 2
        #    while jdx < len(ind[idx]) and n_cent < self.n_neighbors:
        #        diff_vect = (centroid + X_min[ind[idx][jdx]])/(n_cent + 1) - x_min
        #        new_cent_dist = np.sqrt(np.dot(np.dot(diff_vect, metric_tensor), diff_vect))
        #
        #        # checking if new nearest centroid neighbor found
        #        if new_cent_dist < cent_dist:
        #            centroid = centroid + X_min[ind[idx][jdx]]
        #            ncn[idx, n_cent] = ind[idx][jdx]
        #            n_cent = n_cent + 1
        #            cent_dist = new_cent_dist
        #
        #        jdx = jdx + 1
        #
        #    # registering the number of nearest centroid neighbors found
        #    ncn_nums[idx] = n_cent

        # generating samples
        #samples = []
        #while len(samples) < n_to_sample:
        #    random_idx = self.random_state.randint(len(X_min))
        #    random_neighbor_idx = self.random_state.choice(
        #        ncn[random_idx][:ncn_nums[random_idx]])
        #    samples.append(self.sample_between_points(
        #        X_min[random_idx], X_min[random_neighbor_idx]))

        samples = self.generate_samples(X_min, n_to_sample, ncn, ncn_nums)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the MOT2LD.
"""
import numpy as np

import pytest

from smote_variants import MOT2LD

from smote_variants.datasets import load_normal

def test_specific():
    """
    Oversampler specific testing
    """

    with pytest.raises(Exception):
        MOT2LD(d_cut=-1)

    with pytest.raises(Exception):
        MOT2LD(d_cut='dummy')

    obj = MOT2LD()

    with pytest.raises(Exception):
        obj.check_enough_clusters(np.array([0]))

    with pytest.raises(Exception):
        obj.check_enough_peaks(np.array([]))

    with pytest.raises(Exception):
        obj.check_empty_clustering(np.array([[]]))

    obj = MOT2LD(d_cut=2)

    dataset = load_normal()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9]])
    y = np.array([0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0



"""
Testing the CondensedNN.
"""

from smote_variants.noise_removal import CondensedNearestNeighbors

from smote_variants.datasets import (load_all_min_noise,
                                    load_separable)

def test_specific():
    """
    Oversampler specific testing
    """

    obj = CondensedNearestNeighbors()

    dataset = load_all_min_noise()

    X_samp, _ = obj.remove_noise(dataset['data'],
                                 dataset['target'])

    assert len(X_samp) > 0

    dataset = load_separable()

    X_samp, _ = obj.remove_noise(dataset['data'],
                                 dataset['target'])

    assert len(X_samp) > 0



"""
This module implements the SMMO method.
"""
import warnings

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from ..base import instantiate_obj

from .._logger import logger
_logger= logger

__all__= ['SMMO']

class SMMO(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{smmo,
                            author = {de la Calleja, Jorge and Fuentes, Olac
                                        and González, Jesús},
                            booktitle= {Proceedings of the Twenty-First
                                        International Florida Artificial
                                        Intelligence Research Society
                                        Conference},
                            year = {2008},
                            month = {01},
                            pages = {276-281},
                            title = {Selecting Minority Examples from
                                    Misclassified Data for Over-Sampling.}
                            }

    Notes:
        * In this paper the ensemble is not specified. I have selected
            some very fast, basic classifiers.
        * Also, it is not clear what the authors mean by "weighted distance".
        * The original technique is not prepared for the case when no minority
            samples are classified correctly be the ensemble.
    """

    categories = [OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_classifier,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 ensemble=[('sklearn.discriminant_analysis',
                            'QuadraticDiscriminantAnalysis',
                            {}),
                            ('sklearn.tree',
                            'DecisionTreeClassifier',
                            {'random_state': 2}),
                            ('sklearn.naive_bayes',
                            'GaussianNB',
                            {})],
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): control parameter of the nearest neighbor
                                component
            ensemble (list): list of classifiers, if None, default list of
                                classifiers is used
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.ensemble = ensemble
        self.ensemble_objs = [instantiate_obj(e) for e in ensemble]
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        ensembles = [[('sklearn.discriminant_analysis',
                        'QuadraticDiscriminantAnalysis',
                        {}),
                        ('sklearn.tree',
                        'DecisionTreeClassifier',
                        {'random_state': 2}),
                        ('sklearn.naive_bayes',
                        'GaussianNB',
                        {})]]
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'ensemble': ensembles}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self, X_min, ind, n_to_sample):
        """
        Generate the samples.

        Args:
            X_min (np.array): minority samples
            ind (np.array): neighborhood structure
            n_to_sample (int): number of samples to generate

        Returns:
            np.array: the generated samples
        """
        X_means = np.mean(X_min[ind[:, 1:]], axis=1) # pylint: disable=invalid-name
        base_indices = self.random_state.choice(np.arange(X_means.shape[0]),
                                                n_to_sample)
        base_ind, base_counts = np.unique(base_indices, return_counts=True)
        samples = []
        for idx, index in enumerate(base_ind):
            cluster = np.array([X_means[index]])
            indices = np.array([np.hstack([np.array([0]), np.arange(len(cluster))])])
            X_vertices = X_min[ind[index][indices[0]][1:]]
            samples.append(self.sample_simplex(X=cluster,
                                            indices=indices,
                                            n_to_sample=base_counts[idx],
                                            X_vertices=X_vertices))

        samples = np.vstack(samples)
        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # training and in-sample prediction (out-of-sample by k-fold cross
        # validation might be better)
        predictions = []
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            for ens in self.ensemble_objs:
                predictions.append(ens.fit(X, y).predict(X))

        # constructing ensemble prediction
        pred = np.where(np.mean(np.vstack(predictions), axis=0) > 0.5,
                        1, 0)

        # create mask of minority samples to sample
        mask_to_sample = np.where(np.logical_and(pred != y,
                                                 y == self.min_label))[0]
        if len(mask_to_sample) < 2:
            return self.return_copies(X, y, "Not enough minority samples "\
                            f"selected {len(mask_to_sample)}")

        X_min = X[y == self.min_label]
        X_min_to_sample = X[mask_to_sample] # pylint: disable=invalid-name

        # fitting nearest neighbors model for sampling
        n_neighbors = min([len(X_min), self.n_neighbors + 1])

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min_to_sample, return_distance=False)

        samples = self.generate_samples(X_min, ind, n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'ensemble': self.ensemble,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self, deep)}
